/*
 *		l i b l i s t x . h
 *
 *	This is the private header file associated 
 *	with the LIBLIST library.  It contains all
 *	the definitions that are internal to the
 *	implementation of the library (and thus
 *	should never be used by user programs).
 *
 *	The purpose of this library is to provide
 *	an object oriented method of handling list
 *	and tree structures, so that the user does
 *	not have to worry about writing such routines
 *	in all sorts of programs.
 *
 *
 * (c) Copyright 1996-2000 David J. Walker
 *	   Email:  dave.walker@icl.com
 *
 *	   Permission to copy and/or distribute granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they
 *		   arise from defects in it.
 *	   3). Altered version must not be represented as being the
 *		   original software.
 *
 *	AMENDMENT HISTORY
 *	~~~~~~~~~~~~~~~~~
 *	10 May 96	DJW   - First version
 *
 *	24 May 96	DJW   - Added additional definitions needed to build the
 *						debugging version of the LIBLIST library.
 *
 *	28 May 96	DJW   - Added definitions needed to support Hash Tables
 *						and Balanced Binary Trees.
 *					  - Changed existing internal structure definitions to
 *						support the mechanics of the LIST_Clone method.
 *
 *	20 June 96	DJW   - Added data definitions needed to support embedding of
 *						one list in another one.
 *
 *	09 May 99	DJW   - Added definition for _LIST_listRemove support routine.
 *
 *	07 Jul 00	DJW   - Added hash value to _namepace structure.
 *					  - Changed DEBUG macro to be called LISTDEBUG
 */

#ifndef _LIBLISTX_H
#define _LIBLISTX_H

#ifdef WIN32
#ifndef __STDC__
#define __STDC__ 1
#endif
#endif

#ifndef _LIBLIST_H
#include <liblist.h>
#endif /* _LIBLIST_H */

#ifdef WIN32
#include <io.h>
#define write _write
#define read _read
#else
#include <unistd.h>
#endif

#ifndef STDERR
#define STDERR 2
#endif

/*
 *	define the use of the 'assert' macro depending
 *	on whether we are testing the library code or not
 */
#ifdef LIBTEST
 #define LIBASSERT(param)	assert(param);
 #include <assert.h>
#else
 #define LIBASSERT(param)
#endif /* LIBTEST */

#ifdef __STDC__
#define _(params) params
#else
#define _(params) ()
#endif /* __STDC__ */

/*
 *	Special list type definitions that are
 *	used internally by library code, but never
 *	made visible at the user level
 */
#define LIST_CLASS_EMBED	&_LISTembed
#define LIST_TYPE_EMBEDDED 20

/*
 *	Now set up our use of the LIBDEBUG macro
 */
#ifndef LISTDEBUG
	/*
	 *	We define our own specialist use of
	 *	the library routine that supports
	 *	the normal assert functionality
	 */
	#define LIBDEBUG(routine,message,result)
#else
	#undef LIBDEBUG
	/*
	 *	Normal debugging enabled -
	 *	this allows us to do run time checks
	 *	We map this onto the 'assert' mechanism
	 */
	#define _LISTSTR(x) _LISTVAL(x)
	#define _LISTVAL(x) #x
	#define LIBDEBUG(routine,message, expr) ((expr) ? (void)0 \
						:  _LIST_Debug( __FILE__ ":" _LISTSTR(__LINE__) "\n" \
						   routine ":" message));

#endif /* LISTDEBUG */


/*
 *	Master list class definition.
 *	This will be common to all the
 *	derived classes of this generic type.
 */
struct _listmaster
	{
	long	(*count)		_((list_t));
	node_t	(*first)		_((list_t));
	node_t	(*last) 		_((list_t));
	node_t	(*next) 		_((list_t, node_t));
	node_t	(*previous) 	_((list_t, node_t));
	list_t	(*clone)		_((list_t));
	int 	(*destroy)		_((list_t *));
	node_t	(*nodenew)		_((list_t, va_list));
	int 	(*add)			_((list_t, node_t));
	int 	(*after)		_((list_t, node_t, node_t));
	int 	(*remove)		_((list_t, node_t));
	node_t	(*find) 		_((list_t, int (*findfunc)(node_t,va_list), va_list));
	int listtype;		/* List type */
	listclass_t listclass;			/* list class */
#ifdef LISTDEBUG
	char *	listdesc;
#endif /* LISTDEBUG */
	};
typedef struct _listmaster * _listmaster_t;


/*
 *	Generic List/Tree Class Object definition
 *	This is shared amongst all clones of this list
 */
struct _listclass
	{
	struct _listmaster *master; 	/* Master calss pointer */
	size_t	sizelist;				/* Size of the list header */
	size_t	sizenode;				/* Size of the node header */
	size_t	sizeuser;				/* Size of user part of nodes */
	unsigned int	clones; 		/* Clone count */
	int 	duplicates; 	/* Duplicate control */
	int 	(*initnode)_((node_t, va_list));  /* User node Initialise */
	int 	(*killnode)_((node_t)); 		  /* User node Destroy */
	int 	(*compare)_((node_t,node_t));	  /* user node compare */
	};

/*
 *	Generic List/Tree Header
 *	Object definition
 */
struct _listhead
	{
	listclass_t  listclass; 	   /* Pointer to shared class data */
	node_t	nodes;				   /* Pointer to nodes in list */
	};

#ifndef LISTDEBUG
#define _LIST_NODECHECK
#else
typedef struct _nodecheck {
		list_t	list;					/* Instance of list node belongs to */
		listclass_t type;				/* Class of list node belongs to */
		} * _nodecheck_t;

#define _LIST_NODECHECK    struct _nodecheck   check;
#endif /* LISTDEBUG */
/*
 *	Single linked lists definitions
 */
	/*	nodes */
typedef struct _list1
	{
		_LIST_NODECHECK
		struct _list1 * next;
	} * _list1_t;

	/*	header */
typedef struct _list1h
	{
		listclass_t 	listclass;
		_list1_t		nodes;
	} * _list1h_t;


/*
 *	Generic Double linked list definitions
 */
	/* nodes */
typedef struct _list2
	{
		_LIST_NODECHECK
		struct _list2 * next;
		struct _list2 * prev;
	} * _list2_t;

	/* header */
typedef struct _list2h
	{
		listclass_t 	listclass;
		_list2_t		nodes;
		_list2_t		last;
	} * _list2h_t;

/*
 *	Generic FIFO (queue) node definition
 */
	/* nodes */
typedef struct _fifo
	{
		_LIST_NODECHECK
		struct _fifo * next;
	} * _fifo_t;

	/*	header */
	/* (currently same as double linked list header) */
typedef struct _fifoh
	{
		listclass_t 	listclass;
		_fifo_t 		nodes;
		_fifo_t 		last;
	} * _fifoh_t;

/*
 *	Generic LIFO (stack) node definition
 */
	/* nodes */
typedef struct _lifo
	{
		_LIST_NODECHECK
		struct _lifo * next;
	} * _lifo_t;

	/*	header */
	/*	(Currently this is identical to the */
	/*	 single linked list definition) */
typedef struct _lifoh
	{
		listclass_t 	listclass;
		_lifo_t 		nodes;
	} * _lifoh_t;


/*
 *	Generic Binary Tree definitions
 */
	/* nodes */
typedef struct _btree
	{
		_LIST_NODECHECK
		struct	_btree *  up;
		struct	_btree *  left;
		struct	_btree *  right;
	} * _btree_t;
	/*	header */
typedef struct _btreeh
	{
		listclass_t 	listclass;
		_btree_t		nodes;
	} * _btreeh_t;

/*
 *	Generic Balanced Binary Tree definitions
 */
	/* nodes */
typedef struct _baltree
	{
		_LIST_NODECHECK
		struct	_baltree *	up; 		/* \  Note that these three fields */
		struct	_baltree *	left;		/*	> must correspond to the start */
		struct	_baltree *	right;		/* /  of the _btree structure	   */
		size_t	left_depth;
		size_t	right_depth;
	} * _baltree_t;
	/*	header */
typedef struct _baltreeh
	{
		listclass_t 	listclass;
		_baltree_t		nodes;
	} * _baltreeh_t;

/*
 *	Generic Hash table bucket definition
 */
	/* nodes */
typedef struct _hash
	{
		_LIST_NODECHECK
		union {
			list_t	list;
			node_t	node;
			} u;
	} * _hash_t;
	/*	header */
typedef struct _hashh
	{
		listclass_t    listclass;
		_hash_t 	   buckets;
		size_t		   bucketcount; 	/* Number of buckets in table */
		size_t		  (*nodehash) _((node_t));
		size_t		  (*findhash) _((va_list ap));
		list_t		  clonelist;		/* Dummy for cloning */
	} * _hashh_t;

/*
 *	List handle definition for embedded lists.
 */
	/* header */
typedef struct _embedh
	{
		listclass_t 	 listclass; /* Dummy to get vector table */
		list_t			 self;		/* Entry for true list for this level */
		struct _embedh * inner; 	/* List contained in this one (or NULL) */
		struct _embedh * outer; 	/* List containing this one (or NULL) */
	} * _embedh_t;

/*
 *	Structure used to maintain name space
 */
typedef struct _namespace
	{
	char *	name;
	unsigned short hashvalue;
	size_t	count;
	}  * _namespace_t;

/*
 *	Useful macros to convert between the type
 *	of node pointer passed to users which hides
 *	the control structures, and the internal
 *	library variants which do not.
 */
#define _LIST_USERNODE(l,n)  ((void *)((char *)n + ((list_t)l)->listclass->sizenode))
#define _LIST_REALNODE(l,n)  ((void *)((char *)n - ((list_t)l)->listclass->sizenode))

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */
/*
 *	Globally visible variables that
 *	are internal to the implementation
 *	of the LIBLIST library.
 */
extern	struct _listclass _LISTembed;
extern list_t	_LIST_NameSpace;

/*
 *	Globally visible functions that
 *	are internal to the implementation
 *	of the LIBLIST library.
 */
extern list_t  _LIST_CloneClass    _((listclass_t));
extern list_t  _LIST_CloneList	   _((list_t));
extern void    _LIST_UnClone	   _((list_t));
extern int	   _LIST_Compare	   _((list_t, node_t, node_t));
extern int	   _LIST_Destroy	   _((list_t *));
extern node_t  _LIST_NewNode	   _((list_t, va_list));
extern long    _LIST_ListCount	   _((list_t));
extern int	   _LIST_Index		   _((list_t, node_t));
extern _embedh_t _LIST_EmbedOuter  _((list_t));
extern long    _LIST_EmbedCount    _((list_t));
extern node_t  _LIST_EmbedFirst    _((list_t));
extern node_t  _LIST_EmbedLast	   _((list_t));
extern node_t  _LIST_EmbedNext	   _((list_t, node_t));
extern node_t  _LIST_EmbedPrevious _((list_t, node_t));
extern node_t  _LIST_EmbedNew	   _((list_t, va_list));
extern int	   _LIST_EmbedAdd	   _((list_t, node_t));
extern int	   _LIST_EmbedAfter    _((list_t, node_t, node_t));
extern int	   _LIST_EmbedRemove   _((list_t, node_t));
extern node_t  _LIST_EmbedFind	   _((list_t,int (*)(node_t,va_list),va_list));
extern node_t  _LIST_FifoLast	   _((list_t));
extern int	   _LIST_FifoAfter	   _((list_t,node_t,node_t));
extern int	   _LIST_FifoRemove    _((list_t, node_t));
extern node_t  _LIST_LifoLast	   _((list_t));
extern node_t  _LIST_LifoPrevious  _((list_t,node_t));
extern int	   _LIST_LifoAfter	   _((list_t,node_t,node_t));
extern int	   _LIST_LifoRemove    _((list_t, node_t));
extern node_t  _LIST_ListFirst	   _((list_t));
extern node_t  _LIST_ListNext	   _((list_t,node_t));
extern int	   _LIST_ListAdd	   _((list_t,node_t));
extern node_t  _LIST_ListFind	   _((list_t,int (*)(node_t,va_list),va_list));
extern long    _LIST_ListRemove    _((list_t, node_t));
extern node_t  _LIST_NewGetNode    _((list_t, va_list));
extern node_t  _LIST_Remove 	   _((list_t, node_t));
extern node_t  _LIST_BtreeFirst    _((list_t));
extern node_t  _LIST_BtreeLast	   _((list_t));
extern node_t  _LIST_BtreeNext	   _((list_t, node_t));
extern node_t  _LIST_BtreePrevious _((list_t, node_t));
extern int	   _LIST_BtreeRemove   _((list_t, node_t));
extern node_t  _LIST_BtreeFind	   _((list_t, int (*)(node_t, va_list),va_list));
extern node_t  _LIST_HashFirst	   _((list_t));
extern node_t  _LIST_HashLast	   _((list_t));
extern node_t  _LIST_HashPrevious  _((list_t, node_t));
extern node_t  _LIST_HashNext	   _((list_t, node_t));
extern int	   _LIST_HashAfter	   _((list_t, node_t, node_t));
extern int	   _LIST_HashAdd	   _((list_t, node_t));
extern int	   _LIST_HashRemove    _((list_t, node_t));
extern node_t  _LIST_HashFind	   _((list_t, int (*)(node_t, va_list),va_list));

extern void    _LIST_Debug		   _((char *));
extern int	   _LIST_ListCheck	   _((list_t));
extern int	   _LIST_NodeCheck	   _((list_t, node_t));
extern void    _LIST_InitNode	   _((list_t, node_t));

#ifdef LISTTRACK
extern void    _LIST_TrackAdd	   _((list_t));
extern int	   _LIST_TrackDelete   _((list_t));
#endif /* LISTTRACK */

#ifdef __cplusplus
	}
#endif /* __cplusplus */

#endif /* _LIBLISTX_H */
